/*name and version number:@(#)lapm_bg.c	1.120*/
/*date of get: 		  11/13/98 11:33:25*/
/*date of delta:	  08/13/98 17:56:43*/
/****************************************************************
File :	lapm_bg.c
Description :
	Contains all lapm background functions (lapm states).

Procedures Contained :
	lapm_background
	lapm_init
	lapm_disconnect

Revision History :
	Initials	Date		Change
*****************************************************************/

#include "llglob_d.h"
#include "ioglob_d.h"
#include "lapm.h"
#include "mnp.h"
#include "v42bis.h"

#ifdef DSVD
#include "dsvd_d.h"
#ifdef MODEM_VXD
extern void STDCALL VMODEM_Start_Timer ( int );
#endif
#endif

#ifdef CS_MODEM
#include "csglob.h"
#endif

void setup_ETC_modem_rate(void);

/****************************************************************
Routine : lapm_abort
Description :
	Any key abort during lapm negotiation
Parameters in :
	none
Parameters out :
	none
Revision History :
	Initials	Date		Change
*****************************************************************/
void lapm_abort ( void )
{
	// support for any key abort during lapm negotiation
	if ( io_get_dte_rx_char ( &io_dte_rx_last_ch ) )
	{
		x_debug_msg ( "lapm - any key abort" ) ;
		x_status.ud_TerminationCause = UD_TC_ANY_KEY_ABORT;
		lapm_state = LAPM_DOWN ;
		x_send_mail ( V42_REMOTE_DISC, DP_TASK, NULL, NULL ) ;
	}
}
/****************************************************************
Routine : lapm_disconnect
Description :
	send lapm disc command to remote
Parameters in :
	none
Parameters out :
	none
Revision History :
	Initials	Date		Change
*****************************************************************/
void lapm_disconnect ( void )
{
	if ( lapm_state == LAPM_CONNECT )
	{
		lapm_state = LAPM_DISC_CMD ;
		// send multiple disconnect commands - make sure it reaches the remote
		lapm_send_control_frame ( DISC, TRUE, TRUE ) ;
		lapm_send_control_frame ( DISC, TRUE, TRUE ) ;
		lapm_send_control_frame ( DISC, TRUE, TRUE ) ;
		lapm_send_control_frame ( DISC, TRUE, TRUE ) ;
		lapm_reset_timer () ;
	}
}

/****************************************************************
Routine : lapm_init
Description :
	init lapm variables - transmit odp if originator.
Parameters in :
	none
Parameters out :
	none
Revision History :
	Initials	Date		Change
*****************************************************************/
void lapm_init ( void )
{
	mnp_init_variables () ;
	lapm_init_variables () ;
	// switch to async 8 bit, no parity
	x_send_mail ( DP_ASYNC_8BIT_MODE, DP_TASK, 0, 0 ) ;
	if ( x_modem_originate == TRUE )
		lapm_state = LAPM_TX_ODP ;
	else
		lapm_state = LAPM_RX_ODP ;
	lapm_reset_timer () ;
	
#ifdef DSVD
	// if voice call first, then enable dsvd
	if ( ( S.ModemOptions.DsvdEnabled == FALSE ) &&
		 ( S.ModemOptions.DsvdVoiceCallFirst == TRUE ) )
	{
		S.ModemOptions.DsvdEnabled = TRUE ;
		x_debug_msg ( "dsvd - voice call first" ) ;
	}
	else
	{
		S.ModemOptions.DsvdVoiceCallFirst = FALSE ;
	}
	if ( S.ModemOptions.DsvdEnabled == TRUE )
		x_send_mail ( DSVD_IO_TASK, IO_TASK, 0, DSVD_AUTO_CALL ) ;
#endif
}

/****************************************************************
Routine : lapm_v42bis_default
Description :
	Prepare the default values for the values to be negotiated.
Parameters in :
Parameters out :
	none
Revision History :
	Initials	Date		Change
	EE     		7/18/96		Initial
*****************************************************************/
#ifdef GMUX
extern	byte gmux_unerm_data;
extern 	void lapm_v42bis_default ( byte v_n401, byte v_lapm_k, byte v_P0, byte v_P1_hi, byte v_P1_lo, byte v_P2);
void lapm_v42bis_default ( byte v_n401, byte v_lapm_k, byte v_P0, byte v_P1_hi, byte v_P1_lo, byte v_P2)
{
	lapm_n401 =  v_n401;
	lapm_k = v_lapm_k;
	v42bis_P0 = v_P0;	  // enables v42bis_encode, v42bis_decode if selected.
	v42bis_N2 = (v_P1_hi << 8) + v_P1_lo;
	v42bis_N7 = v_P2;
	v42bis_init_variables () ;
#ifdef DEBUGGING
	sprintf ( szBuffer, "n401 :%d", lapm_n401) ;
	x_debug_msg ( szBuffer ) ;
	sprintf ( szBuffer, "lapm_k: %d, v42bis_P0 :%d", 
							lapm_k, v42bis_P0) ;
	x_debug_msg ( szBuffer ) ;
	sprintf ( szBuffer, "v42bis_N2: %d, v42bis_N7 :%d", 
							v42bis_N2, v42bis_N7) ;
	x_debug_msg ( szBuffer ) ;
#endif
}
#endif

/****************************************************************
Routine : lapm_background
Description :
	lapm background state process
Parameters in :
	none
Parameters out :
	TRUE - stay in lapm state 
	FALSE - lapm init failed, return to bufferd or mnp mode
Revision History :
	Initials	Date		Change
*****************************************************************/
byte lapm_background ( byte new_state )
{
	byte frame_type ;

	if ( new_state != LAPM_KEEP_STATE )
		lapm_state = new_state ;

	switch ( lapm_state )
	{
  		case LAPM_IDLE :
			// nothing to do, just check if any key abort
			lapm_abort () ;
			break;

		case LAPM_TX_ODP :
			if ( lapm_adp_detected () == TRUE )
			{
				// clear transmit fifo
				io_init_dce_tx_fifo () ;	// init dce transmit fifo
				// tell dp to switch to hdlc mode
				x_send_mail ( DP_HDLC_MODE, DP_TASK, 0, 0 ) ;
				lapm_reset_timer () ;
				lapm_state = LAPM_TX_HDLC_FLAGS ;
			}
			else
			{
				if ( ( ETC_mode ) || ( S.ModemOptions.V42 == V42_LAPM_ONLY )
						|| ( S.ModemOptions.Enabled56K == TRUE ) )
				{
					// try to connect for atleast 30 seconds
					if ( lapm_timer_expired ( SECOND30 ) == TRUE )
					{
						x_debug_msg ( "LAPM - adp detect timer expired" ) ;
#ifdef CS_MODEM
						disconnect_reason = V42_FAIL;
#endif
						return ( LAPM_TO_BUFFERED_MODE ) ;		// return to bufferd mode
					}
				}
				// if timer expired return to mnp or buffered mode
				else if ( (lapm_timer_expired ( SECOND2 ) == TRUE) )
				{
						x_debug_msg ( "LAPM - adp detect timer expired" ) ;
						if ( S.ModemOptions.V42 == V42_LAPM_BUFFER )
							return ( LAPM_TO_BUFFERED_MODE ) ;
						lapm_state = LAPM_IDLE ;
						mnp_init() ;
						io_state = IO_MNP ;	// try mnp mode
						return ( lapm_state ) ;
				}
				if ( io_dce_tx_fifo_empty() )
					lapm_transmit_odp () ;
			}
			lapm_abort () ;
			break ;

		case LAPM_TX_HDLC_FLAGS :
			// transmit flags for at least 300 msec
			if ( lapm_timer_expired ( MS300 ) == TRUE )
			{
				x_debug_msg ( "LAPM - sent xid cmd" ) ;
				lapm_send_xid ( TRUE ) ;	/* set command flag */
				lapm_reset_timer () ;
				lapm_retransmit = 0 ;
				lapm_state = LAPM_TX_XID_CMD ;
				if ( S.ModemOptions.Enabled56K == TRUE )
				{
#if !defined(CS_MODEM)
					// update 'ati11' diagnostics information
					dp_update_diagnostics () ;
					x_status.IniTransmitRate = x_line_rate_tx ;
					x_status.IniReceiveRate = x_line_rate ;
#endif
				}
			}
			lapm_abort () ;
			break ;

		case LAPM_TX_XID_CMD :
			lapm_rx_hdlc_frame () ;
			frame_type = lapm_rx_frame_type () ;
			if ( frame_type == XID )
			{
				x_debug_msg ( "LAPM - received xid rsp" ) ;
				lapm_process_xid () ;
				lapm_send_control_frame ( SABME, TRUE, TRUE ) ;
				lapm_reset_timer () ;
				lapm_retransmit = 0 ;
				lapm_state = LAPM_TX_SABME_CMD ;
			}
			else if ( lapm_timer_expired ( lapm_t401 ) == TRUE )
			{
				if ( ( ETC_mode ) || ( S.ModemOptions.V42 == V42_LAPM_ONLY )
						|| ( S.ModemOptions.Enabled56K == TRUE ) )
				{
					if ( lapm_retransmit >= 10 )
					{
						x_debug_msg("LAPM - no response for xid cmd");
						return ( LAPM_TO_BUFFERED_MODE ) ;
					}
				}
				else if ( lapm_retransmit >= 3 )
				{
					x_debug_msg("LAPM - no response for xid cmd");
					return ( LAPM_TO_BUFFERED_MODE ) ;	// return to buffered mode
				}
				// one more try of xid command
				++lapm_retransmit ;
				x_debug_msg ( "LAPM - sent xid cmd" ) ;
				lapm_send_xid ( TRUE ) ;	/* set command flag */
				lapm_reset_timer () ;
			}
			if ( frame_type == IFRM )	/* clear if i-frame */
				lapm_clear_rx_frame () ;
			lapm_abort () ;
			break ;

		case LAPM_TX_SABME_CMD :
			lapm_rx_hdlc_frame () ;
			frame_type = lapm_rx_frame_type () ;
			if ( ( frame_type == UA ) ||
				 ( frame_type == IFRM ) )
			{
				lapm_state = LAPM_PRE_CONNECT ;
			}
			else if ( lapm_timer_expired ( lapm_t401 ) == TRUE )
			{
				if ( ( ETC_mode ) || ( S.ModemOptions.V42 == V42_LAPM_ONLY )
						|| ( S.ModemOptions.Enabled56K == TRUE ) )
				{
					if ( lapm_retransmit >= 10 )
					{
						x_debug_msg ( "LAPM - no response for sabme cmd" ) ;
						return ( LAPM_TO_BUFFERED_MODE ) ;
					}
				}
				else if ( lapm_retransmit >= 3 )
				{
					x_debug_msg ( "LAPM - no response for sabme cmd" ) ;
					return ( LAPM_TO_BUFFERED_MODE ) ;	// return to buffered mode
				}
				// one more try of sabme command
				++lapm_retransmit ;
				lapm_send_control_frame ( SABME, TRUE, TRUE ) ;
				lapm_reset_timer () ;
			}
			lapm_abort () ;
			break ;

		case LAPM_RX_ODP :
			if ( ETC_mode )
			{
				// send hdlc flags - skip odp detection phase
				lapm_reset_timer () ;
				// tell dp to switch to hdlc mode
				x_send_mail ( DP_HDLC_MODE, DP_TASK, 0, 0 ) ;
				// do no send adp - receive xid and sabme
				lapm_state = LAPM_RX_SABME_CMD ;
			}
			else if ( lapm_odp_detected () == TRUE )
			{
				x_debug_msg ( "LAPM - odp detected" ) ;
				lapm_transmit_adp () ;
				lapm_reset_timer () ;
				lapm_state = LAPM_TX_ADP ;
			}
			else if ( ( S.ModemOptions.V42 != V42_LAPM_BUFFER ) &&
					  ( S.ModemOptions.V42 != V42_LAPM_ONLY ) &&
					  ( lapm_mnp_lr_detected () == TRUE ) )
			{
				lapm_state = LAPM_IDLE ;
				io_state = IO_MNP ;
				mnp_state = MNP_ANSW_INITIAL ;
			}
			// if timer expired return to bufferd mode - No EC
			else if ( ( lapm_timer_expired ( SECOND5 ) == TRUE ) &&
					  ( S.ModemOptions.V42 != V42_LAPM_ONLY ) )
			{
#ifdef CS_MODEM
				disconnect_reason = V42_FAIL;
#endif
				x_debug_msg ( "LAPM - odp detect timer expired" ) ;
				return ( LAPM_TO_BUFFERED_MODE ) ;	// return to bufferd mode
			}
			lapm_abort () ;
			break ;

		case LAPM_TX_ADP :
			if ( ( io_dce_tx_fifo_empty() ) ||
				 ( lapm_timer_expired ( MS100 ) == TRUE ) )
			{
				lapm_state = LAPM_RX_XID_CMD ;
				// tell dp to switch to hdlc mode
				x_send_mail ( DP_HDLC_MODE, DP_TASK, 0, 0 ) ;
			}
			lapm_abort () ;
			break ;

		case LAPM_RX_XID_CMD :
			lapm_rx_hdlc_frame () ;
			frame_type = lapm_rx_frame_type () ;
			if ( frame_type == XID )
			{
				x_debug_msg ( "LAPM - received xid cmd" ) ;
				// clear transmit fifo - the adp sequence still in fifo
				io_init_dce_tx_fifo () ;	// init dce transmit fifo
				lapm_process_xid () ;
				lapm_reset_timer () ;
				lapm_state = LAPM_RX_SABME_CMD ;
			}
			else if ( frame_type == SABME )
			{
				lapm_send_control_frame ( UA, FALSE, TRUE ) ;
				lapm_reset_timer () ;
				lapm_state = LAPM_PRE_CONNECT ;
			}
			else
			{
				// try to connect for atleast 5 seconds
				if ( lapm_timer_expired ( SECOND5 ) == TRUE )
				{
#ifdef CS_MODEM
					disconnect_reason = LAPM_TIMEOUT;
#endif
					x_debug_msg ( "LAPM - xid detect timer expired" ) ;
					return ( LAPM_TO_BUFFERED_MODE ) ;	// return to bufferd mode
				}
				else if ( lapm_timer_expired ( SECOND2 ) == TRUE )
				{
					/* keep transmitting ADP to remote 
					   - for a reliable connection */
					lapm_transmit_adp () ;
				}
			}
			if ( frame_type == IFRM )	/* clear if i-frame */
				lapm_clear_rx_frame () ;
			lapm_abort () ;
			break ;

		case LAPM_RX_SABME_CMD :
			/* transmit the xid frame - if fifo was full */
			lapm_tx_hdlc_frame () ;
			lapm_rx_hdlc_frame () ;
			frame_type = lapm_rx_frame_type () ;
			if ( frame_type == SABME )
			{
				lapm_send_control_frame ( UA, FALSE, TRUE ) ;
				lapm_reset_timer () ;
				lapm_state = LAPM_PRE_CONNECT ;
			}
			else if ( frame_type == XID )
			{
				x_debug_msg ( "LAPM - received xid cmd" ) ;
				lapm_process_xid () ;
				lapm_reset_timer () ;
				/* stay in this state for sabme command - if we got
				an xid in this state then the remote did not receive
				our previous xid response */
			}
			else
			{
				if ( ( ETC_mode ) || ( S.ModemOptions.Enabled56K == TRUE ) )
				{
					if ( lapm_timer_expired ( SECOND30 ) == TRUE )
					{
#ifdef CS_MODEM
						disconnect_reason = LAPM_TIMEOUT;
#endif
						x_debug_msg ( "LAPM - xid/sabme detect timer expired" ) ;
						return ( LAPM_TO_BUFFERED_MODE ) ;	// return to bufferd mode
					}
				}
				// try to connect for atleast 5 seconds
				else if ( lapm_timer_expired ( SECOND5 ) == TRUE )
				{
#ifdef CS_MODEM
					disconnect_reason = LAPM_TIMEOUT;
#endif
					x_debug_msg ( "LAPM - sabme detect timer expired" ) ;
					return ( LAPM_TO_BUFFERED_MODE ) ;	// return to bufferd mode
				}
			}
			if ( frame_type == IFRM )	// clear if i-frame
				lapm_clear_rx_frame () ;
			lapm_abort () ;
			break ;

		case LAPM_DISC_CMD :
			lapm_rx_hdlc_frame () ;
			frame_type = lapm_rx_frame_type () ;
			if ( ( frame_type == UA ) || ( frame_type == DISC ) ||
				( lapm_timer_expired ( lapm_t401 ) == TRUE ) ) /*+ #ETC ***/
			{
				/* the remote acknowledged our disc command - or timer
				expired - send cleardown sequence and then send OK */
				lapm_state = LAPM_DOWN ;
				x_send_mail ( V42_LOCAL_DISC, DP_TASK, NULL, NULL ) ;
			}
			if ( lapm_receive_frame.status == FULL ) // clear frames
				lapm_clear_rx_frame () ;
			break ;

		case LAPM_DISC_RSP :
			if ( lapm_timer_expired ( MS100 ) == TRUE )
			{
				/* the remote commanded us to disconnect - send cleardown
				sequence then send NO CARRIER */
				lapm_state = LAPM_DOWN ;
				x_send_mail ( V42_REMOTE_DISC, DP_TASK, NULL, NULL ) ;
			}
			break ;

		case LAPM_DOWN :
			/* do not do anything in this state - wait for response
			from dp task after the clear down sequence - io will receive
			a disconnect from dp task - switch back to io idle state */
			break ;

		case LAPM_PRE_CONNECT :
			lapm_state = LAPM_CONNECT_INIT ;
#ifdef DSVD
			if ( S.ModemOptions.DsvdEnabled )
				lapm_state = LAPM_DSVD_CONNECT ;
#endif
			lapm_abort () ;
			break ;

#ifdef DSVD
		case LAPM_DSVD_CONNECT :
			lapm_rx_hdlc_frame () ;		// receive new frame
			if ( x_dsvd_state == DSVD_DATA_ONLY )
			{
				x_dsvd_connect = TRUE ;			// DSVD connection
				lapm_state = LAPM_CONNECT_INIT ;
#ifdef MODEM_VXD
				// put the timer to 5 secs for dos box dsvd initiated
				// connection
				VMODEM_Start_Timer ( 5 ) ;
#endif
			}
			else if ( x_dsvd_state == DSVD_INIT )
			{
				x_dsvd_connect = FALSE ;		// no DSVD connection
				lapm_state = LAPM_CONNECT_INIT ;
			}
			// clear I-frame
			frame_type = lapm_rx_frame_type () ;
			if ( frame_type == IFRM )
				lapm_clear_rx_frame () ;
			lapm_abort () ;
			break ;
#endif

		case LAPM_CONNECT_INIT :
			lapm_state = LAPM_CONNECT ;
			if ( v42bis_P0 )
				x_modem_ec = EC_V42B ;
#ifdef GMUX
			else if	(gmux_unerm_data)
				x_modem_ec = EC_NO;	// No error corection.
#endif
			else
				x_modem_ec = EC_V42 ;
			x_debug_msg ( "LAPM - to connect state" ) ;
			x_send_mail ( IO_AT_CONNECT, AT_TASK, 0, 0 ) ;
			setup_ETC_modem_rate () ;
			lapm_abort () ;
			break ;

		case LAPM_CONNECT :
			lapm_rx_hdlc_frame () ;		// receive new frame
			/* validate new receive frame - set lapm_poll flag -
			also process ack iframe sequence number N(R) */
			frame_type = lapm_rx_frame_type () ;

			switch ( frame_type )
			{
				case IFRM :
					if ( io_at_esc_detect == FALSE )
						lapm_process_iframe () ;
					lapm_rx_iframe = TRUE ;
					lapm_clear_rx_frame () ;
					break ;

				case RR :
					/* if remote was busy - then continue to transmit */
					if ( lapm_remote_busy == TRUE )
						lapm_reset_timer () ;		/* reset timer T401 */
					lapm_remote_busy = FALSE ;
					break ;

				case RNR :
					lapm_reset_timer () ;		/* reset timer T401 */
					lapm_remote_busy = TRUE ;
					// do not retransmit if remote is busy
					lapm_retransmit = 0 ;
					break ;

				case REJ :
					lapm_ns = lapm_va ;		/* retransmit */
					lapm_reset_timer () ;		/* reset timer T401 */
					lapm_remote_busy = FALSE ;

#if !defined(CS_MODEM)
					x_status.TxErrorCount++;
#endif
					break ;

				case XID :
					lapm_process_xid () ;
					break ;

				case UI :
					lapm_process_break () ;
					break ;

				case DISC :
					lapm_state = LAPM_DISC_RSP ;
#ifdef CS_MODEM
					disconnect_reason = LAPMDISC;
#endif
					x_status.ud_TerminationCause = UD_TC_RX_DISC_FRAME;
					// send multiple ack's to make sure it reaches the remote
					lapm_send_control_frame ( UA, FALSE, TRUE ) ;
					lapm_send_control_frame ( UA, FALSE, TRUE ) ;
					lapm_send_control_frame ( UA, FALSE, TRUE ) ;
					lapm_send_control_frame ( UA, FALSE, TRUE ) ;
					lapm_reset_timer () ;
					break ;

				case SABME :
					lapm_send_control_frame ( UA, FALSE, TRUE ) ;
					break ;
			}

			if ( io_at_esc_detect == FALSE )
			{
				// process data in receive i-frame's temp buffer
				lapm_process_iframe () ;
				// build new i-frames
				lapm_build_iframe () ;
			}
			// transmit any control frame or i-frame (if remote not busy)
			lapm_tx_hdlc_frame () ;

			// if remote polled (or) i-frames received - send an ack
			if ( lapm_poll ||
				 (lapm_rx_iframe && (io_get_dce_queue_count () == 0)) )
			{
				if ( ( lapm_local_busy ) || ( io_at_esc_detect ) )
					lapm_send_control_frame ( RNR, FALSE, lapm_poll ) ;
				else
					lapm_send_control_frame ( RR, FALSE, lapm_poll ) ;
				lapm_poll = FALSE ;
				lapm_rx_iframe = FALSE ;
			}

			if ( lapm_timer_expired ( lapm_t401 ) == TRUE )
			{
				lapm_reset_timer () ;		/* reset timer T401 */
				lapm_rx_iframe = FALSE ;
				/* if idle, poll the remote for an acknowledgment */
				if ( ( lapm_local_busy ) || ( io_at_esc_detect ) )
					lapm_send_control_frame ( RNR, TRUE, TRUE ) ;
				else
					lapm_send_control_frame ( RR, TRUE, TRUE ) ;
				/* if remote has not acknowledged i-frames then retransmit.
				   if T401 expired - the timer is reset if a i-frame
				   is available to transmit - also the timer is reset
				   if a i-frame is acknowledged N(R) */
				if ( lapm_ns != lapm_va )
				{
					x_debug_msg ( "LAPM - ack timer expired" ) ;
					++lapm_retransmit ;
					++x_status.ud_RetransmitCount;
					// if retransmissions exceeded N400 count then disconnect
					if ( lapm_retransmit >= N400 )
					{
#ifdef CS_MODEM
						disconnect_reason = LAPM_RETRY_LIMIT;
#endif
						lapm_state = LAPM_DISC_RSP ;
						x_status.ud_TerminationCause = UD_TC_N400_TIME_OUT;
						x_debug_msg ( "LAPM - n400 timeout disconnect" ) ;
					}
					else if ( lapm_remote_busy == FALSE )
					{
						x_debug_msg ( "LAPM - retransmit" ) ;
						lapm_ns = lapm_va ;		// retransmit
						// retrain the dsp if timeout count exceeded
						if ( lapm_retransmit >= N400/2 )
							x_send_mail(IO_DP_RETRAIN, DP_TASK, NULL, NULL);
					}
				}
			}
			break ;

		case LAPM_TO_BUFFERED_MODE :
		case LAPM_KEEP_STATE :
			// virtual states, nothing to do...
			break;
	}
	return ( lapm_state ) ;		/* stay in v42 state */
}

/****************************************************************
Routine : setup_ETC_modem_rate
Description :
	update modem rate after successful ETC negotiation
Parameters in :
	none
Parameters out :
	none
Revision History :
	Initials	Date		Change
*****************************************************************/
void setup_ETC_modem_rate ( void )
{
	if ( ETC_mode )
	{
		x_send_mail ( UPDATE_MODEM_RATE, DP_TASK, NULL, NULL ) ;
	}
}

extern void UART_stop_break ( void ) ;

/****************************************************************
Routine : lapm_tx_break
Description :
	send lapm break command to remote
Parameters in :
	none
Parameters out :
	none
Revision History :
	Initials	Date		Change
*****************************************************************/
void lapm_tx_break ( void )
{
	lapm_send_break ( TRUE ) ;	/* set command/response param to true */
	UART_stop_break () ;
}
